import ast
import io
import logging
import tokenize
import warnings

import numpy as np

try:
    from ast import unparse
except ImportError:
    from astunparse import unparse as _unparse

    def unparse(*args):
        return _unparse(*args).strip("\n")


logger = logging.getLogger("sharrow.aster")


def unparse_(*args):
    # spit out a ast dump on error
    try:
        return unparse(*args)
    except Exception:
        logger.warning(ast.dump(*args))
        raise


ast_Constant_Type = ast.Constant


def ast_String_value(x):
    return x.value if isinstance(x, ast.Str) else x


ast_TupleIndex_Type = ast.Tuple


def ast_Index_Value(x):
    return x


ast_Constant = ast.Constant


def _isNone(c):
    if c is None:
        return True
    if isinstance(c, ast_Constant_Type) and c.value is None:
        return True
    return False


def extract_all_name_tokens(command):
    all_names = set()
    if isinstance(command, str):
        command = command.encode()
    for token in tokenize.tokenize(io.BytesIO(command).readline):
        if token.type == tokenize.NAME:
            all_names.add(token.string)
    return all_names


def extract_names(command):
    names = set()
    z = ast.parse(command.lstrip())
    for node in ast.walk(z):
        if isinstance(node, ast.Name):
            names.add(node.id)
    return names


def extract_names_2(command):
    if not isinstance(command, str):
        return set(), dict(), dict()

    z = ast.parse(command.lstrip())

    # Attribute-Name pairs first
    attribute_name_pairs = dict()
    skip_names = set()
    for i in ast.walk(z):
        if isinstance(i, ast.Attribute):
            if not isinstance(i.value, ast.Name):
                continue
            k = i.value.id
            a = i.attr
            skip_names.add(id(i.value))
            if k not in attribute_name_pairs:
                attribute_name_pairs[k] = set()
            attribute_name_pairs[k].add(a)

    # Subscript pairs next
    subscript_name_pairs = dict()
    for i in ast.walk(z):
        if isinstance(i, ast.Subscript):
            if not isinstance(i.value, ast.Name):
                continue
            k = i.value.id
            try:
                a = i.slice.value.value  # Python <= 3.8
            except AttributeError:
                try:
                    a = (
                        i.slice.value
                    )  # Python == 3.9  # FIXME when dropping support for 3.8
                except AttributeError:
                    a = None
            skip_names.add(id(i.value))
            if a is not None:
                if k not in subscript_name_pairs:
                    subscript_name_pairs[k] = set()
                subscript_name_pairs[k].add(a)

    # Plain names last
    plain_names = set()
    for i in ast.walk(z):
        if isinstance(i, ast.Name):
            if id(i) in skip_names:
                continue
            plain_names.add(i.id)

    return plain_names, attribute_name_pairs, subscript_name_pairs


class RewriteExpression(ast.NodeTransformer):
    def visit_Call(self, node):
        try:
            tag = node.func.id
        except AttributeError:
            tag = None
        if tag == "clip":
            if len(node.args) == 3:
                basic, lower, upper = node.args
            elif len(node.args) == 2:
                basic, lower = node.args
                upper = None
            else:
                raise ValueError("incorrect number of args for clip")
            if _isNone(upper):
                return ast.Call(
                    func=ast.Name(id="where", ctx=ast.Load()),
                    args=[
                        ast.Compare(
                            left=basic,
                            ops=[ast.Gt()],
                            comparators=[lower],
                        ),
                        basic,
                        lower,
                    ],
                    keywords=[],
                )
            elif _isNone(lower):
                return ast.Call(
                    func=ast.Name(id="where", ctx=ast.Load()),
                    args=[
                        ast.Compare(
                            left=basic,
                            ops=[ast.Lt()],
                            comparators=[upper],
                        ),
                        basic,
                        upper,
                    ],
                    keywords=[],
                )
            else:
                return ast.Call(
                    func=ast.Name(id="where", ctx=ast.Load()),
                    args=[
                        ast.Compare(
                            left=basic,
                            ops=[ast.Gt()],
                            comparators=[lower],
                        ),
                        ast.Call(
                            func=ast.Name(id="where", ctx=ast.Load()),
                            args=[
                                ast.Compare(
                                    left=basic,
                                    ops=[ast.Lt()],
                                    comparators=[upper],
                                ),
                                basic,
                                upper,
                            ],
                            keywords=[],
                        ),
                        lower,
                    ],
                    keywords=[],
                )
        elif tag == "piece":
            if len(node.args) == 3:
                basic, lower, upper = node.args
            elif len(node.args) == 2:
                basic, lower = node.args
                upper = None
            else:
                raise ValueError("incorrect number of args for piece")
            if _isNone(upper):
                clip = ast.Call(
                    func=ast.Name(id="where", ctx=ast.Load()),
                    args=[
                        ast.Compare(
                            left=basic,
                            ops=[ast.Gt()],
                            comparators=[lower],
                        ),
                        basic,
                        lower,
                    ],
                    keywords=[],
                )
                return ast.BinOp(
                    left=clip,
                    op=ast.Sub(),
                    right=lower,
                )
            elif _isNone(lower):
                return ast.Call(
                    func=ast.Name(id="where", ctx=ast.Load()),
                    args=[
                        ast.Compare(
                            left=basic,
                            ops=[ast.Lt()],
                            comparators=[upper],
                        ),
                        basic,
                        upper,
                    ],
                    keywords=[],
                )
            else:
                clip = ast.Call(
                    func=ast.Name(id="where", ctx=ast.Load()),
                    args=[
                        ast.Compare(
                            left=basic,
                            ops=[ast.Gt()],
                            comparators=[lower],
                        ),
                        ast.Call(
                            func=ast.Name(id="where", ctx=ast.Load()),
                            args=[
                                ast.Compare(
                                    left=basic,
                                    ops=[ast.Lt()],
                                    comparators=[upper],
                                ),
                                basic,
                                upper,
                            ],
                            keywords=[],
                        ),
                        lower,
                    ],
                    keywords=[],
                )
                return ast.BinOp(
                    left=clip,
                    op=ast.Sub(),
                    right=lower,
                )
        elif tag == "max":
            if len(node.args) == 2:
                left, right = node.args
            else:
                raise ValueError(
                    "incorrect number of args for max (currently only 2 is allowed)"
                )
            return ast.Call(
                func=ast.Name(id="where", ctx=ast.Load()),
                args=[
                    ast.Compare(
                        left=left,
                        ops=[ast.Lt()],
                        comparators=[right],
                    ),
                    right,
                    left,
                ],
                keywords=[],
            )
        elif tag == "min":
            if len(node.args) == 2:
                left, right = node.args
            else:
                raise ValueError(
                    "incorrect number of args for max (currently only 2 is allowed)"
                )
            return ast.Call(
                func=ast.Name(id="where", ctx=ast.Load()),
                args=[
                    ast.Compare(
                        left=left,
                        ops=[ast.Lt()],
                        comparators=[right],
                    ),
                    left,
                    right,
                ],
                keywords=[],
            )
        else:
            return node


def bool_wrap(subnode):
    if (
        isinstance(subnode, ast.Call)
        and isinstance(subnode.func, ast.Attribute)
        and subnode.func.attr == "bool_"
        and isinstance(subnode.func.value, ast.Name)
        and subnode.func.value.id == "np"
    ):
        return subnode
    return ast.Call(
        func=ast.Attribute(
            ast.Name("np"),
            "bool_",
        ),
        args=[subnode],
        keywords={},
    )


class RewriteForNumba(ast.NodeTransformer):
    def __init__(
        self,
        spacename,
        dim_slots,
        spacevars=None,
        rawname="_inputs",
        rawalias="____",
        digital_encodings=None,
        preferred_spacename=None,
        extra_vars=None,
        blenders=None,
        bool_wrapping=False,
        swallow_errors=False,
        get_default=False,
        original_expr="???",
    ):
        self.spacename = spacename
        self.dim_slots = dim_slots
        self.spacevars = spacevars
        self.rawname = rawname
        self.rawalias = rawalias
        self.digital_encodings = digital_encodings or {}
        self.preferred_spacename = preferred_spacename
        self.extra_vars = extra_vars or {}
        self.blenders = blenders or {}
        self.bool_wrapping = bool_wrapping
        self.swallow_errors = swallow_errors
        self.get_default = get_default
        self.original_expr = original_expr

    def log_event(self, tag, node1=None, node2=None):
        if logger.getEffectiveLevel() <= 0:
            if node1 is None:
                logger.debug(f"RewriteForNumba({self.spacename}|{self.rawalias}).{tag}")
            elif node2 is None:
                try:
                    unparsed = unparse_(node1)
                except:  # noqa: E722
                    unparsed = f"{type(node1)} not unparseable"
                logger.debug(
                    f"RewriteForNumba({self.spacename}|{self.rawalias}).{tag} "
                    f"[{type(node1).__name__}]= {unparsed}",
                )
            else:
                try:
                    unparsed1 = unparse_(node1)
                except:  # noqa: E722
                    unparsed1 = f"{type(node1).__name__} not unparseable"
                try:
                    unparsed2 = unparse_(node2)
                except:  # noqa: E722
                    unparsed2 = f"{type(node2).__name__} not unparseable"
                logger.debug(
                    f"RewriteForNumba({self.spacename}|{self.rawalias}).{tag} "
                    f"[{type(node1).__name__},{type(node2).__name__}]= "
                    f"{unparsed1} => {unparsed2}",
                )

    def generic_visit(self, node):
        self.log_event("generic_visit", node)
        return super().generic_visit(node)

    def _replacement(
        self,
        attr,
        ctx,
        original_node,
        topname=None,
        transpose_lead=False,
        missing_dim_value=None,
    ):
        if topname is None:
            topname = self.spacename
        pref_topname = self.preferred_spacename or topname

        if self.spacevars is not None:
            if attr not in self.spacevars:
                if self.get_default or (
                    topname == pref_topname and not self.swallow_errors
                ):
                    raise KeyError(
                        f"{topname}..{attr}\nexpression={self.original_expr}"
                    )
                # we originally raised a KeyError here regardless, but what if
                # we just give back the original node, and see if other spaces,
                # possibly fallback spaces, might work?  If nothing works then
                # it will still eventually error out when compiling?
                # The swallow errors option allows us to continue processing
                # using the same rules, then circle back later to clean up
                # errors, so that as-yet unprocessed Ast elements get the
                # chance to be seen by the first pass.
                return original_node

        dim_slots = self.dim_slots
        if isinstance(self.spacevars, dict):
            dim_slots = self.spacevars[attr]

        def maybe_transpose(thing):
            if transpose_lead:
                return ast.Call(
                    func=ast.Name("transpose_leading", ctx=ast.Load()),
                    args=[thing],
                    keywords=[],
                )
            else:
                return thing

        def _maybe_transpose_first_two_args(_slice):
            if transpose_lead:
                elts = _slice.elts
                if len(elts) >= 2:
                    elts = [elts[1], elts[0], *elts[2:]]
                return type(_slice)(elts=elts)
            else:
                return _slice

        raw_decorated_name = f"__{pref_topname}__{attr}"
        decorated_name = maybe_transpose(
            ast.Name(id=raw_decorated_name, ctx=ast.Load())
        )
        logger.debug(f"    decorated_name= {unparse_(decorated_name)}")

        if isinstance(dim_slots, (tuple, list)):
            if len(dim_slots):
                elts = []
                for n in dim_slots:
                    if isinstance(n, int):
                        elts.append(ast.Name(id=f"_arg{n:02}", ctx=ast.Load()))
                    elif isinstance(n, dict):
                        elts.append(
                            ast.Constant(n=n[missing_dim_value], ctx=ast.Load())
                        )
                    else:
                        elts.append(n)
                    logger.debug(f"ELT {unparse_(elts[-1])}")
                s = ast.Tuple(
                    elts=elts,
                    ctx=ast.Load(),
                )
                result = ast.Subscript(
                    value=decorated_name,
                    slice=s,
                    ctx=ctx,
                )
                logger.debug(f"two+ dim_slots on decorated_name {unparse_(result)}")
            else:
                # no indexing, just name replacement
                result = decorated_name
                logger.debug(f"just decorated_name {unparse_(result)}")
        else:
            if isinstance(dim_slots, int):
                s = ast.Name(id=f"_arg{dim_slots:02}", ctx=ast.Load())
            else:
                s = dim_slots
            result = ast.Subscript(
                value=decorated_name,
                slice=s,
                ctx=ctx,
            )
            logger.debug(f"one dim_slots on decorated_name {unparse_(result)}")

        digital_encoding = self.digital_encodings.get(attr, None)
        if digital_encoding is not None:
            dictionary = digital_encoding.get("dictionary", None)
            offset_source = digital_encoding.get("offset_source", None)
            if dictionary is not None:
                result = ast.Subscript(
                    value=ast.Name(
                        id=f"__encoding_dict__{pref_topname}__{attr}", ctx=ast.Load()
                    ),
                    slice=result,
                    ctx=ctx,
                )
            elif offset_source is not None:
                result = ast.Subscript(
                    value=maybe_transpose(
                        ast.Name(
                            id=f"__{pref_topname}__{offset_source}", ctx=ast.Load()
                        )
                    ),
                    slice=result.slice,
                    ctx=ctx,
                )
                result = ast.Subscript(
                    value=ast.Name(id=f"__{pref_topname}__{attr}", ctx=ast.Load()),
                    slice=result,
                    ctx=ctx,
                )
            else:
                missing_value = digital_encoding.get("missing_value", None)
                if missing_value is not None:
                    scale = digital_encoding.get("scale", 1)
                    offset = digital_encoding.get("offset", 0)
                    result = ast.Call(
                        func=ast.Name("digital_decode", cts=ast.Load()),
                        args=[
                            result,
                            ast.Num(scale),
                            ast.Num(offset),
                            ast.Num(missing_value),
                        ],
                        keywords=[],
                    )
                else:
                    scale = digital_encoding.get("scale", 1)
                    offset = digital_encoding.get("offset", 0)
                    if scale != 1:
                        result = ast.BinOp(
                            left=result,
                            op=ast.Mult(),
                            right=ast.Num(scale),
                        )
                    if offset:
                        result = ast.BinOp(
                            left=result,
                            op=ast.Add(),
                            right=ast.Num(offset),
                        )

        blender = self.blenders.get(attr, None)
        if blender is not None:
            # get_blended_2(backstop, indices, indptr, data, i, j, blend_limit=np.inf)
            result_args = result.slice.elts
            # inside the blender, the args will be maz-taz mapped, but we need the plain (i)maz too now
            result_arg_ = [j.slice for j in result_args]
            if len(result_args) == 2:
                result = ast.Call(
                    func=ast.Name("get_blended_2", cts=ast.Load()),
                    args=[
                        result,
                        ast.Name(
                            id=f"__{pref_topname}___s_{attr}__indices", ctx=ast.Load()
                        ),
                        ast.Name(
                            id=f"__{pref_topname}___s_{attr}__indptr", ctx=ast.Load()
                        ),
                        ast.Name(
                            id=f"__{pref_topname}___s_{attr}__data", ctx=ast.Load()
                        ),
                        result_arg_[0 if not transpose_lead else 1],
                        result_arg_[1 if not transpose_lead else 0],
                        ast_Constant(blender.get("max_blend_distance")),  # blend_limit
                    ],
                    keywords=[],
                )
            else:
                raise NotImplementedError()

        self.log_event(f"_replacement({attr}, {topname})", original_node, result)
        return result

    def visit_Subscript(self, node):
        if isinstance(node.value, ast.Name):
            # for XXX[YYY], XXX is a space name and YYY is a literal value: skims['DIST']
            if (
                node.value.id == self.spacename
                and isinstance(node.slice, ast_Constant_Type)
                and isinstance(ast_String_value(node.slice), str)
            ):
                self.log_event(f"visit_Subscript(Constant {node.slice.value})")
                return self._replacement(ast_String_value(node.slice), node.ctx, node)
            # for XXX[YYY], XXX is the raw placeholds and YYY is a literal value: ____['income']
            if (
                node.value.id == self.rawalias
                and isinstance(node.slice, ast_Constant_Type)
                and isinstance(ast_String_value(node.slice), str)
                and ast_String_value(node.slice) in self.spacevars
            ):
                result = ast.Subscript(
                    value=ast.Name(id=self.rawname, ctx=ast.Load()),
                    slice=ast.Constant(self.spacevars[ast_String_value(node.slice)]),
                    ctx=node.ctx,
                )
                self.log_event(
                    f"visit_Subscript(Raw {ast_String_value(node.slice)})",
                    node,
                    result,
                )
                return result
            # for XXX[YYY,ZZZ], XXX is a space name and YYY is a literal value and
            # ZZZ is a literal value: skims['SOV_TIME','MD']
            if (
                node.value.id == self.spacename
                and isinstance(ast_Index_Value(node.slice), ast.Tuple)
                and len(ast_Index_Value(node.slice).elts) == 2
                and isinstance(ast_Index_Value(node.slice).elts[0], ast_Constant_Type)
                and isinstance(ast_Index_Value(node.slice).elts[1], ast_Constant_Type)
                and isinstance(
                    ast_String_value(ast_Index_Value(node.slice).elts[0]), str
                )
                and isinstance(
                    ast_String_value(ast_Index_Value(node.slice).elts[1]), str
                )
            ):
                _a = ast_String_value(ast_Index_Value(node.slice).elts[0])
                _b = ast_String_value(ast_Index_Value(node.slice).elts[1])
                self.log_event(f"visit_Subscript(Tuple ({_a}, {_b}))")
                return self._replacement(
                    _a,
                    node.ctx,
                    node,
                    missing_dim_value=_b,
                )
            # for XXX[...], there is no space name and XXX is the name of an aux_var
            if (
                node.value.id in self.spacevars
                and isinstance(self.spacevars[node.value.id], ast.Name)
                and self.spacename == ""
            ):
                result = ast.Subscript(
                    value=self.spacevars[node.value.id],
                    slice=self.visit(node.slice),
                    ctx=node.ctx,
                )
                self.log_event(
                    f"visit_Subscript(AuxVar {node.value.id})",
                    node,
                    result,
                )
                return result
        self.log_event("visit_Subscript(no change)", node)
        return node

    def visit_Attribute(self, node):
        if isinstance(node.value, ast.Name):
            # for XXX.YYY, XXX is a space name and YYY is a literal value: skims.DIST
            if node.value.id == self.spacename:
                return self._replacement(node.attr, node.ctx, node)
            # for ____.YYY, handles unadorned values in the top level
            if node.value.id == self.rawalias and node.attr in self.spacevars:
                result = ast.Subscript(
                    value=ast.Name(id=self.rawname, ctx=ast.Load()),
                    slice=ast.Constant(self.spacevars[node.attr]),
                    ctx=node.ctx,
                )
                self.log_event(f"visit_Attribute(Raw {node.attr})", node, result)
                return result
            # for YYY.ZZZ, where YYY is a variable in the root and ZZZ is anything
            if self.spacename == "" and node.value.id in self.spacevars:
                result = ast.Attribute(
                    value=self.visit(node.value),
                    attr=node.attr,
                    ctx=node.ctx,
                )
                self.log_event("visit_Attribute(lead change)", node, result)
                return result
            return node
        else:
            # pass through
            result = ast.Attribute(
                value=self.visit(node.value),
                attr=node.attr,
                ctx=node.ctx,
            )
            self.log_event("visit_Attribute(no change)", node, result)
            return result

    def visit_Name(self, node):
        attr = node.id
        if attr not in self.spacevars:
            self.log_event("visit_Name(no change)", node)
            return node
        if self.spacename == "":
            if isinstance(self.spacevars[attr], ast.Name):
                # when spacevars values are ast.Name we are using it, it's probably an aux_var
                result = self.spacevars[attr]
            else:
                result = ast.Subscript(
                    value=ast.Name(id=self.rawname, ctx=ast.Load()),
                    slice=ast.Constant(self.spacevars[attr]),
                    ctx=node.ctx,
                )
            self.log_event(f"visit_Name(Constant {attr})", node, result)
            return result
        else:
            result = self._replacement(attr, node.ctx, node, self.spacename)
            self.log_event(f"visit_Name(Replacement {attr})", node, result)
            return result

    def visit_UnaryOp(self, node):
        # convert bitflip `~x` operator into `~np.bool_(x)`
        if self.bool_wrapping and isinstance(node.op, ast.Invert):
            return ast.UnaryOp(
                op=node.op,
                operand=bool_wrap(self.visit(node.operand)),
            )
        else:
            return ast.UnaryOp(op=node.op, operand=self.visit(node.operand))

    def visit_BinOp(self, node):
        # convert bitwise binops:
        # `x & y` -> `np.bool_(x) & np.bool_(y)`
        # `x | y` -> `np.bool_(x) | np.bool_(y)`
        # `x ^ y` -> `np.bool_(x) ^ np.bool_(y)`
        left = self.visit(node.left)
        right = self.visit(node.right)

        if self.bool_wrapping and isinstance(
            node.op, (ast.BitAnd, ast.BitOr, ast.BitXor)
        ):
            result = ast.BinOp(
                left=bool_wrap(left),
                op=node.op,
                right=bool_wrap(right),
            )
            self.log_event("visit_BinOp(Replacement)", node, result)
        else:
            result = ast.BinOp(
                left=left,
                op=node.op,
                right=right,
            )
            self.log_event("visit_BinOp(no change)", node, result)
        return result

    def visit_Call(self, node):
        result = None
        # implement ActivitySim's "reverse" skims
        if (
            isinstance(node.func, ast.Attribute) and node.func.attr == "reverse"
        ):  # *.reverse(...)
            if isinstance(node.func.value, ast.Name):  # somename.reverse(...)
                if node.func.value.id == self.spacename:  # spacename.reverse(...)
                    if len(node.args) == 1 and isinstance(
                        node.args[0], ast_Constant_Type
                    ):  # spacename.reverse('constant')
                        result = self._replacement(
                            ast_String_value(node.args[0]),
                            node.func.ctx,
                            None,
                            transpose_lead=True,
                        )
        # handle clip as a method
        if isinstance(node.func, ast.Attribute) and node.func.attr == "clip":
            if isinstance(node.func.value, ast.Name) and node.func.value.id == "np":
                # call to np.clip(...), change to local clip implementation
                clip_args = []
                for a in node.args:
                    clip_args.append(self.visit(a))
                result = ast.Call(
                    func=ast.Name("clip", cts=ast.Load()),
                    args=clip_args,
                    keywords=[self.visit(i) for i in node.keywords],
                )
            elif len(node.args) == 1 and len(node.keywords) == 0:
                # single positional arg becomes max
                result = ast.Call(
                    func=ast.Name("max", cts=ast.Load()),
                    args=[self.visit(node.func.value), self.visit(node.args[0])],
                    keywords=[],
                )
            elif len(node.args) == 0 and len(node.keywords) == 1:
                # single keyword argument, what is it?
                if node.keywords[0].arg in ("upper", "max"):
                    # becomes min
                    result = ast.Call(
                        func=ast.Name("min", cts=ast.Load()),
                        args=[
                            self.visit(node.func.value),
                            self.visit(node.keywords[0].value),
                        ],
                        keywords=[],
                    )
                elif node.keywords[0].arg in ("lower", "min"):
                    # becomes max
                    result = ast.Call(
                        func=ast.Name("max", cts=ast.Load()),
                        args=[
                            self.visit(node.func.value),
                            self.visit(node.keywords[0].value),
                        ],
                        keywords=[],
                    )
            elif len(node.args) == 2 and len(node.keywords) == 0:
                # two positional arg becomes max and min
                nested_call = ast.Call(
                    func=ast.Name("min", cts=ast.Load()),
                    args=[self.visit(node.func.value), self.visit(node.args[1])],
                    keywords=[],
                )
                result = ast.Call(
                    func=ast.Name("max", cts=ast.Load()),
                    args=[nested_call, self.visit(node.args[0])],
                    keywords=[],
                )
            else:
                # two-way clip with keywords
                # move clip when used as a class method to be a regular function
                # we provide a basic implementation of clip in sharrow.maths
                clip_args = [self.visit(node.func.value)]
                for a in node.args:
                    clip_args.append(self.visit(a))
                result = ast.Call(
                    func=ast.Name("clip", cts=ast.Load()),
                    args=clip_args,
                    keywords=[self.visit(i) for i in node.keywords],
                )
        # move x.apply(func, **kwargs) to be func(x, **kwargs)
        if isinstance(node.func, ast.Attribute) and node.func.attr == "apply":
            apply_args = [self.visit(node.func.value)]
            assert len(node.args) == 1
            apply_func = self.visit(node.args[0])
            result = ast.Call(
                func=apply_func,
                args=apply_args,
                keywords=[self.visit(i) for i in node.keywords],
            )
        # implement ActivitySim's "max" skims
        if isinstance(node.func, ast.Attribute) and node.func.attr == "max":
            if isinstance(node.func.value, ast.Name):
                if node.func.value.id == self.spacename:
                    if len(node.args) == 1 and isinstance(
                        node.args[0], ast_Constant_Type
                    ):
                        forward = self._replacement(
                            ast_String_value(node.args[0]), node.func.ctx, None
                        )
                        backward = self._replacement(
                            ast_String_value(node.args[0]),
                            node.func.ctx,
                            None,
                            transpose_lead=True,
                        )
                        result = ast.Call(
                            func=ast.Name("max", ctx=ast.Load()),
                            # func=ast.Attribute(
                            #     value=ast.Name("np", ctx=ast.Load()),
                            #     attr="maximum",
                            # ),
                            args=[forward, backward],
                            keywords=[],
                        )
        # change `x.astype(int, **kwargs)` to `int(x, **kwargs)`
        if isinstance(node.func, ast.Attribute) and node.func.attr == "astype":
            apply_args = [self.visit(node.func.value)]
            assert len(node.args) == 1
            apply_func = self.visit(node.args[0])
            result = ast.Call(
                func=apply_func,
                args=apply_args,
                keywords=[self.visit(i) for i in node.keywords],
            )
        # change `x.isin([2,3,4])` to `x == 2 or x == 3 or x == 4`
        if isinstance(node.func, ast.Attribute) and node.func.attr == "isin":
            ante = self.visit(node.func.value)
            targets = node.args
            elts = None
            if len(targets) == 1 and isinstance(targets[0], (ast.List, ast.Tuple)):
                elts = targets[0].elts
            elif len(targets) == 1 and isinstance(targets[0], ast.Name):
                extra_val = self.extra_vars.get(targets[0].id, None)
                if isinstance(extra_val, (list, tuple)):
                    elts = [ast_Constant(i) for i in extra_val]
            if elts is not None:
                ors = []
                for elt in elts:
                    ors.append(
                        ast.Compare(
                            left=ante, ops=[ast.Eq()], comparators=[self.visit(elt)]
                        )
                    )
                result = self.visit(ast.BoolOp(op=ast.Or(), values=ors))
        # change `x.between(a,b)` to `(a <= x) & (x <= b)`
        if isinstance(node.func, ast.Attribute) and node.func.attr == "between":
            ante = self.visit(node.func.value)
            targets = node.args
            if len(targets) == 2:
                lb, ub = targets
                left = ast.Compare(
                    left=ante, ops=[ast.GtE()], comparators=[self.visit(lb)]
                )
                right = ast.Compare(
                    left=ante, ops=[ast.LtE()], comparators=[self.visit(ub)]
                )
                result = ast.BinOp(left=left, op=ast.BitAnd(), right=right)

        # change XXX.isna() [with no arguments] to np.isnan(x)
        if (
            isinstance(node.func, ast.Attribute)
            and node.func.attr == "isna"
            and len(node.args) == 0
            and len(node.keywords) == 0
        ):
            apply_args = [self.visit(node.func.value)]  # convert XXX into argument
            result = ast.Call(
                func=ast.Name("isnan_fast_safe"),
                args=apply_args,
                keywords=[],
            )
            # if the value XXX is a categorical, rather than just `np.isnan` we also
            # want to check if the category number is less than zero
            if isinstance(apply_args[0], ast.Subscript):
                if isinstance(apply_args[0].value, ast.Name):
                    if apply_args[0].value.id.startswith("__encoding_dict__"):
                        cat_is_lt_zero = ast.Compare(
                            left=apply_args[0].slice,
                            ops=[ast.Lt()],
                            comparators=[ast.Num(0)],
                        )
                        result = ast.BoolOp(
                            op=ast.Or(), values=[cat_is_lt_zero, result]
                        )

        # implement x.get("y",z) where x is the spacename
        if (
            isinstance(node.func, ast.Attribute)
            and node.func.attr == "get"
            and node.func.value.id == self.spacename
        ):
            if len(node.args) == 2 and len(node.keywords) == 0:
                try:
                    result = self._replacement(
                        ast_String_value(node.args[0]), node.func.value.ctx, node
                    )
                except KeyError:
                    if self.get_default:
                        result = self.visit(node.args[1])
                    else:
                        raise
            if (
                len(node.args) == 1
                and len(node.keywords) == 1
                and "default" == node.keywords[0].arg
            ):
                try:
                    result = self._replacement(
                        ast_String_value(node.args[0]), node.func.value.ctx, node
                    )
                except KeyError:
                    if self.get_default:
                        result = self.visit(node.keywords[0].value)
                    else:
                        raise
            if len(node.args) == 1 and len(node.keywords) == 0:
                result = self._replacement(
                    ast_String_value(node.args[0]), node.func.value.ctx, node
                )

        # change np.where(x, y, z) to (y if x else z), for performance reasons
        if (
            isinstance(node.func, ast.Attribute)
            and node.func.attr == "where"
            and isinstance(node.func.value, ast.Name)
            and (node.func.value.id == "np" or node.func.value.id == "numpy")
            and len(node.args) == 3
        ):
            if len(node.args) == 3 and len(node.keywords) == 0:
                result = ast.IfExp(
                    test=self.visit(node.args[0]),
                    body=self.visit(node.args[1]),
                    orelse=self.visit(node.args[2]),
                )

        # if no other changes
        if result is None:
            args = [self.visit(i) for i in node.args]
            kwds = [self.visit(i) for i in node.keywords]
            result = ast.Call(
                func=self.visit(node.func),
                args=args,
                keywords=kwds,
            )
        self.log_event("visit_Call", node, result)
        return result

    def visit_Compare(self, node):
        result = None

        # devolve XXX ==/!= YYY when XXX or YYY is a categorical and the other is a constant category value
        if len(node.ops) == 1 and isinstance(node.ops[0], (ast.Eq, ast.NotEq)):
            left = self.visit(node.left)
            right = self.visit(node.comparators[0])
            left_is_categorical = (
                isinstance(left, ast.Subscript)
                and isinstance(left.value, ast.Name)
                and left.value.id.startswith("__encoding_dict__")
            )
            if left_is_categorical and isinstance(right, ast.Constant):
                # left is categorical, right is a constant
                left_spacename = left.value.id.split("__")[2]
                left_varname = left.value.id.split("__")[3]
                if (
                    left_spacename == self.spacename
                    and left_varname in self.digital_encodings
                ):
                    left_dictionary = self.digital_encodings[left_varname].get(
                        "dictionary", np.atleast_1d([])
                    )
                    try:
                        right_decoded = np.where(left_dictionary == right.value)[0][0]
                    except IndexError:
                        right_decoded = None
                        warnings.warn(
                            f"right hand value {right.value!r} not found in "
                            f"categories for {left_varname} in {self.spacename}"
                            f"\nexpression: {self.original_expr}"
                            f"\ncategories: {left_dictionary}",
                            stacklevel=2,
                        )
                        # at this point, the right value is not in the left's categories, so
                        # it is guaranteed to be not equal to any of the categories.
                        if isinstance(node.ops[0], ast.Eq):
                            result = ast.Constant(False)
                        elif isinstance(node.ops[0], ast.NotEq):
                            result = ast.Constant(True)
                        else:
                            raise ValueError(
                                f"unexpected operator {node.ops[0]}"
                            ) from None
                    if right_decoded is not None:
                        result = ast.Compare(
                            left=left.slice,
                            ops=[self.visit(i) for i in node.ops],
                            comparators=[ast_Constant(right_decoded)],
                        )
            right_is_categorical = (
                isinstance(right, ast.Subscript)
                and isinstance(right.value, ast.Name)
                and right.value.id.startswith("__encoding_dict__")
            )
            if right_is_categorical and isinstance(left, ast.Constant):
                # right is categorical, left is a constant
                right_spacename = right.value.id.split("__")[2]
                right_varname = right.value.id.split("__")[3]
                if (
                    right_spacename == self.spacename
                    and right_varname in self.digital_encodings
                ):
                    right_dictionary = self.digital_encodings[right_varname].get(
                        "dictionary", np.atleast_1d([])
                    )
                    try:
                        left_decoded = np.where(right_dictionary == left.value)[0][0]
                    except IndexError:
                        left_decoded = None
                        warnings.warn(
                            f"left hand value {left.value!r} not found in "
                            f"categories for {right_varname} in {self.spacename}"
                            f"\nexpression: {self.original_expr}"
                            f"\ncategories: {right_dictionary}",
                            stacklevel=2,
                        )
                        # at this point, the left value is not in the right's categories, so
                        # it is guaranteed to be not equal to any of the categories.
                        if isinstance(node.ops[0], ast.Eq):
                            result = ast.Constant(False)
                        elif isinstance(node.ops[0], ast.NotEq):
                            result = ast.Constant(True)
                        else:
                            raise ValueError(
                                f"unexpected operator {node.ops[0]}"
                            ) from None
                    if left_decoded is not None:
                        result = ast.Compare(
                            left=ast_Constant(left_decoded),
                            ops=[self.visit(i) for i in node.ops],
                            comparators=[right.slice],
                        )

        # if no other changes
        if result is None:
            result = ast.Compare(
                left=self.visit(node.left),
                ops=[self.visit(i) for i in node.ops],
                comparators=[self.visit(i) for i in node.comparators],
            )
        self.log_event("visit_Compare", node, result)
        return result


def expression_for_numba(
    expr,
    spacename,
    dim_slots,
    spacevars=None,
    rawname="_inputs",
    rawalias="____",
    digital_encodings=None,
    prefer_name=None,
    extra_vars=None,
    blenders=None,
    bool_wrapping=False,
    swallow_errors=False,
    get_default=False,
    original_expr=None,
):
    """
    Rewrite an expression so numba can compile it.

    Parameters
    ----------
    expr : str
        The expression being rewritten
    spacename : str
        A namespace of variables that might be in the expression.
    dim_slots : tuple or Any
    spacevars : Mapping, optional
    rawname : str
    rawalias : str
    digital_encodings : Mapping, optional
    prefer_name : str, optional
    extra_vars : Mapping, optional
    blenders : Mapping, optional
    bool_wrapping : bool, optional
    swallow_errors : bool, optional
    get_default : bool, optional
    original_expr : str, optional
        Original (pre-processing) expression, used for debugging

    Returns
    -------
    str
    """
    with warnings.catch_warnings(record=True) as warning_list:
        result = unparse_(
            RewriteForNumba(
                spacename,
                dim_slots,
                spacevars,
                rawname,
                rawalias,
                digital_encodings,
                prefer_name,
                extra_vars,
                blenders,
                bool_wrapping,
                swallow_errors,
                get_default,
                original_expr=original_expr or expr,
            ).visit(ast.parse(expr))
        )
    for warning in warning_list:
        warnings.warn(warning.message, warning.category, stacklevel=2)
    return result


class Asterize:
    def __init__(self):
        self._cache = {}

    def __call__(self, expr):
        target, result = self._cache.get(expr, (None, None))
        if result is None:
            tree = ast.parse(expr, mode="exec")
            new_tree = ast.fix_missing_locations(RewriteExpression().visit(tree))
            if isinstance(new_tree.body[0], ast.Assign):
                a = new_tree.body[0]
                if not len(a.targets) == 1 or not isinstance(a.targets[0], ast.Name):
                    raise ValueError(
                        f"only one simple named assignment target can be given: {expr}"
                    )
                target = a.targets[0].id
                new_tree = a.value
            result = unparse_(new_tree)
            self._cache[expr] = (target, result)
        return target, result


def SimpleSplitter(x):
    if "=" in x:
        s = x.split("=")
        if len(s) == 2:
            return s[0].strip(), s[1].strip()
        else:
            raise ValueError(
                f"only one simple named assignment target can be given: {x}"
            )
    else:
        return None, x



import os.path
from glob import glob

import nbformat as nbf

# Collect a list of all notebooks in the docs folder
notebooks = glob(
    os.path.join(os.path.dirname(__file__), "..", "**", "*.ipynb"),
    recursive=True,
)

# Text to look for in adding tags
text_search_dict = {
    "# TEST": "remove-cell",  # Remove the whole cell
    "# HIDDEN": "hide-cell",  # Hide the whole cell
    "# NO CODE": "remove-input",  # Remove only the input
    "# HIDE CODE": "hide-input",  # Hide the input w/ a button to show
}

# Search through each notebook and look for th text, add a tag if necessary
for ipath in notebooks:
    if "/_build/" in ipath:
        continue
    touch = False
    ntbk = nbf.read(ipath, nbf.NO_CONVERT)

    for cell in ntbk.cells:
        cell_tags = cell.get("metadata", {}).get("tags", [])
        for key, val in text_search_dict.items():
            if key in cell["source"]:
                if val not in cell_tags:
                    cell_tags.append(val)
                    touch = True
        if len(cell_tags) > 0:
            cell["metadata"]["tags"] = cell_tags
    if 1:
        print(f"hiding test cells in {ipath}")
        with open(ipath, "w", encoding="utf-8") as f:
            nbf.write(ntbk, f)
    else:
        print(f"no changes in {ipath}")



def example(extract='m', estimate=False):
    import pandas as pd

    import larch as lx
    from larch import P, X

    m = lx.example(1)
    m.title = "MTC Example 30 (Constrained Simple MNL)"
    m.compute_engine = "numba"

    m_explicit = m.copy()

    m_explicit.utility_ca = P.tottime * X.tottime + P.tottime * 3 * X.totcost
    m_explicit.remove_unused_parameters()
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result_explicit = m_explicit.maximize_loglike(stderr=True)

    from larch.model.constraints import RatioBound

    m.pmaximum = {"totcost": 0, "tottime": 0}

    m.constraints = [
        RatioBound("totcost", "tottime", min_ratio=3.0, max_ratio=999.0, scale=100),
    ]

    result = m.maximize_loglike(stderr=True)

    result

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            "totcost",
            "tottime",
        ),
        (
            "ASCs",
            "ASC.*",
        ),
        (
            "Income",
            "hhinc.*",
        ),
    )

    m.parameter_summary()
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from collections.abc import MutableSequence

import numpy as np
import pandas as pd
from numba import float32, float64, njit
from scipy.optimize import LinearConstraint


class ParametricConstraint(ABC):
    def __init__(self, binding_tol=1e-4):
        self.binding_tol = binding_tol

    def is_binding(self, x):
        return np.absolute(self.fun(x)) < self.binding_tol

    @abstractmethod
    def get_parameters(self):
        raise NotImplementedError("abstract base class, use a derived class instead")

    @abstractmethod
    def get_binding_note(self, x):
        raise NotImplementedError("abstract base class, use a derived class instead")

    @abstractmethod
    def fun(self, x):
        raise NotImplementedError("abstract base class, use a derived class instead")

    @abstractmethod
    def jac(self, x):
        raise NotImplementedError("abstract base class, use a derived class instead")

    @abstractmethod
    def link_model(self, model, scale=None):
        raise NotImplementedError("abstract base class, use a derived class instead")

    def as_constraint_dicts(self):
        return [
            dict(type="ineq", fun=self.fun, jac=self.jac),
        ]


class RatioBound(ParametricConstraint):
    def __init__(
        self,
        p_num,
        p_den,
        min_ratio=None,
        max_ratio=None,
        model=None,
        scale=10,
        binding_tol=1e-4,
    ):
        self.i_num = 0
        self.i_den = 0
        self.cmin_num = 0.0
        self.cmin_den = 0.0
        self.cmax_num = 0.0
        self.cmax_den = 0.0
        super().__init__(binding_tol=binding_tol)
        self.cmax_num = 0
        self.cmax_den = 0
        self.p_num = p_num
        self.p_den = p_den
        self.min_ratio = min_ratio
        self.max_ratio = max_ratio
        self.scale = 1
        self.link_model(model, scale)

    def __repr__(self):
        return f"larch.RatioBound({self.p_num},{self.p_den},{self.min_ratio},{self.max_ratio})"

    def link_model(self, model, scale=None):
        if scale is not None:
            self.scale = scale
        if model is not None:
            self.i_num = model.get_param_loc(self.p_num)
            self.i_den = model.get_param_loc(self.p_den)
            self.len = model.n_params
            if self.min_ratio is not None:
                scaling = max(self.min_ratio, 1 / self.min_ratio) * self.scale
                if model.get_value(self.p_den, kind="minimum") >= 0:
                    # positive denominator
                    self.cmin_num = 1 * scaling
                    self.cmin_den = -self.min_ratio * scaling
                elif model.get_value(self.p_den, kind="maximum") <= 0:
                    # negative denominator
                    self.cmin_num = -1 * scaling
                    self.cmin_den = self.min_ratio * scaling
                else:
                    raise ValueError(
                        "denominator must be bounded to be non-positive or non-negative"
                    )
            else:
                self.cmin_num = 0
                self.cmin_den = 0
            if self.max_ratio is not None:
                scaling = max(self.max_ratio, 1 / self.max_ratio) * self.scale
                if model.get_value(self.p_den, kind="minimum") >= 0:
                    # positive denominator
                    self.cmax_num = -1 * scaling
                    self.cmax_den = self.max_ratio * scaling
                elif model.get_value(self.p_den, kind="maximum") <= 0:
                    # negative denominator
                    self.cmax_num = 1 * scaling
                    self.cmax_den = -self.max_ratio * scaling
                else:
                    raise ValueError(
                        "denominator must be bounded to be non-positive or non-negative"
                    )
            else:
                self.cmax_num = 0
                self.cmax_den = 0

    def _min_fun(self, x):
        return x[self.i_num] * self.cmin_num + x[self.i_den] * self.cmin_den

    def _max_fun(self, x):
        return x[self.i_num] * self.cmax_num + x[self.i_den] * self.cmax_den

    def fun(self, x):
        return min(
            self._min_fun(x),
            self._max_fun(x),
        )

    def jac(self, x):
        j = np.zeros_like(x)
        if self._min_fun(x) < self._max_fun(x):
            j[self.i_num] = self.cmin_num
            j[self.i_den] = self.cmin_den
        else:
            j[self.i_num] = self.cmax_num
            j[self.i_den] = self.cmax_den
        return j

    def get_parameters(self):
        return (self.p_num, self.p_den)

    def get_binding_note(self, x):
        if self._min_fun(x) < self._max_fun(x):
            return f"{self.p_num!s} / {self.p_den!s} ≥ {self.min_ratio}"
        else:
            return f"{self.p_num!s} / {self.p_den!s} ≤ {self.max_ratio}"

    def as_linear_constraints(self):
        a = np.zeros([2, self.len], dtype="float64")
        a[0, self.i_num] = self.cmin_num
        a[0, self.i_den] = self.cmin_den
        a[1, self.i_num] = self.cmax_num
        a[1, self.i_den] = self.cmax_den
        return [LinearConstraint(a, 0, np.inf)]

    def as_soft_penalty(self):
        from .numbamodel import d_softplus, softplus  # here to avoid circular imports

        i_num = self.i_num
        i_den = self.i_den
        cmin_num = self.cmin_num
        cmin_den = self.cmin_den
        cmax_num = self.cmax_num
        cmax_den = self.cmax_den
        scale = self.scale

        @njit(
            [
                float32(float32[:], float32, float32),
                float64(float64[:], float64, float64),
            ]
        )
        def penalty(x, intensity, sharpness=1.0):
            _min = x[i_num] * cmin_num + x[i_den] * cmin_den
            _max = x[i_num] * cmax_num + x[i_den] * cmax_den
            return -softplus(-np.minimum(_min, _max) * scale * intensity, sharpness)

        @njit(
            [
                float32[:](float32[:], float32, float32),
                float64[:](float64[:], float64, float64),
            ]
        )
        def dpenalty(x, intensity, sharpness=1.0):
            j = np.zeros_like(x)
            _min = x[i_num] * cmin_num + x[i_den] * cmin_den
            _max = x[i_num] * cmax_num + x[i_den] * cmax_den
            partial = (
                d_softplus(-np.minimum(_min, _max), sharpness * scale * intensity)
                * scale
                * intensity
            )
            if _min < _max:
                j[i_num] = cmin_num * partial
                j[i_den] = cmin_den * partial
            else:
                j[i_num] = cmax_num * partial
                j[i_den] = cmax_den * partial
            return j

        @njit(
            [
                float32[:](float32[:], float32),
                float64[:](float64[:], float64),
            ]
        )
        def dpenalty_money(x, intensity):
            j = np.zeros_like(x)
            _min = x[i_num] * cmin_num + x[i_den] * cmin_den
            _max = x[i_num] * cmax_num + x[i_den] * cmax_den
            if np.absolute(_min) < 1e-5:
                partial = 0.5 * scale * intensity
                j[i_num] = cmin_num * partial
                j[i_den] = cmin_den * partial
            elif np.absolute(_max) < 1e-5:
                partial = 0.5 * scale * intensity
                j[i_num] = cmax_num * partial
                j[i_den] = cmax_den * partial
            return j

        return penalty, dpenalty, dpenalty_money

    def __eq__(self, other):
        if not isinstance(other, RatioBound):
            return False
        if (
            (self.p_num == other.p_num)
            and (self.p_den == other.p_den)
            and (self.min_ratio == other.min_ratio)
            and (self.max_ratio == other.max_ratio)
        ):
            return True
        return False


class OrderingBound(ParametricConstraint):
    def __init__(self, p_less, p_more=None, model=None, scale=10, binding_tol=1e-4):
        # default initializers
        self.i_less = 0
        self.i_more = 0
        if p_more is None:
            if "<=" in p_less:
                p_less, p_more = (_.strip() for _ in p_less.split("<=", 1))
            elif "<" in p_less:
                p_less, p_more = (_.strip() for _ in p_less.split("<", 1))
            elif ">=" in p_less:
                p_more, p_less = (_.strip() for _ in p_less.split(">=", 1))
            elif ">" in p_less:
                p_more, p_less = (_.strip() for _ in p_less.split(">", 1))
            else:
                raise ValueError(f"cannot interpret {p_less} as ordering")
        super().__init__(binding_tol=binding_tol)
        self.p_less = p_less
        self.p_more = p_more
        self.scale = 1
        if model is not None:
            self.len = model.n_params
            self.link_model(model, scale)
        else:
            self.len = 0

    def __repr__(self):
        return f"larch.OrderingBound({self.p_less},{self.p_more})"

    def link_model(self, model, scale=None):
        self.i_less = model.get_param_loc(self.p_less)
        self.i_more = model.get_param_loc(self.p_more)
        self.len = model.n_params
        if scale is not None:
            self.scale = scale

    def fun(self, x):
        return (x[self.i_more] - x[self.i_less]) * self.scale

    def jac(self, x):
        j = np.zeros_like(x)
        j[self.i_more] = self.scale
        j[self.i_less] = -self.scale
        return j

    def get_parameters(self):
        return (self.p_less, self.p_more)

    def get_binding_note(self, x):
        return f"{self.p_less!s} ≤ {self.p_more!s}"

    def as_linear_constraints(self):
        a = np.zeros([1, self.len], dtype="float64")
        a[0, self.i_more] = self.scale
        a[0, self.i_less] = -self.scale
        return [LinearConstraint(a, 0, np.inf)]

    def as_soft_penalty(self):
        from .numbamodel import d_softplus, softplus  # here to avoid circular imports

        i_more = self.i_more
        i_less = self.i_less
        scale = self.scale

        @njit(
            [
                float32(float32[:], float32, float32),
                float64(float64[:], float64, float64),
            ]
        )
        def penalty(x, intensity, sharpness=1.0):
            return -softplus(-(x[i_more] - x[i_less]) * scale * intensity, sharpness)

        @njit(
            [
                float32[:](float32[:], float32, float32),
                float64[:](float64[:], float64, float64),
            ]
        )
        def dpenalty(x, intensity, sharpness=1.0):
            j = np.zeros_like(x)
            partial = (
                d_softplus(-(x[i_more] - x[i_less]), sharpness * scale * intensity)
                * scale
                * intensity
            )
            j[i_more] = partial
            j[i_less] = -partial
            return j

        @njit(
            [
                float32[:](float32[:], float32),
                float64[:](float64[:], float64),
            ]
        )
        def dpenalty_bind(x, intensity):
            j = np.zeros_like(x)
            diff = x[i_more] - x[i_less]
            if (
                (x[i_more] + x[i_less]) == 0
                or np.absolute(diff / (x[i_more] + x[i_less])) < 5e-3
                or np.absolute(diff) < 1e-6
            ):
                partial = 0.5 * scale * intensity
                j[i_more] = partial
                j[i_less] = -partial
            return j

        return penalty, dpenalty, dpenalty_bind

    def __eq__(self, other):
        if not isinstance(other, OrderingBound):
            return False
        if (self.p_less == other.p_less) and (self.p_more == other.p_more):
            return True
        return False


class FixedBound(ParametricConstraint):
    def __init__(
        self, p, minimum=None, maximum=None, model=None, scale=10, binding_tol=1e-4
    ):
        super().__init__(binding_tol=binding_tol)
        self.p = p
        self.minimum = minimum
        self.maximum = maximum
        self.scale = 1
        if model is not None:
            self.link_model(model, scale)

    def __repr__(self):
        return f"larch.FixedBound({self.p},{self.minimum},{self.maximum})"

    def link_model(self, model, scale=None):
        self.i = model.get_param_loc(self.p)
        self.len = model.n_params
        if scale is not None:
            self.scale = scale

    def _min_fun(self, x):
        if self.minimum is not None:
            if isinstance(x, pd.Series):
                return (x.iloc[self.i] - self.minimum) * self.scale
            else:
                return (x[self.i] - self.minimum) * self.scale
        else:
            return np.inf

    def _max_fun(self, x):
        if self.maximum is not None:
            if isinstance(x, pd.Series):
                return (self.maximum - x.iloc[self.i]) * self.scale
            else:
                return (self.maximum - x[self.i]) * self.scale
        else:
            return np.inf

    def fun(self, x):
        return min(
            self._min_fun(x),
            self._max_fun(x),
        )

    def jac(self, x):
        j = np.zeros_like(x)
        if self._min_fun(x) < self._max_fun(x):
            j[self.i] = self.scale
        else:
            j[self.i] = -self.scale
        return j

    def get_parameters(self):
        return (self.p,)

    def get_binding_note(self, x):
        if self._min_fun(x) < self._max_fun(x):
            return f"{self.p!s} ≥ {self.minimum}"
        else:
            return f"{self.p!s} ≤ {self.maximum}"

    def as_linear_constraints(self):
        a = np.zeros([1, self.len], dtype="float64")
        a[0, self.i] = 1
        return [
            LinearConstraint(
                a,
                self.minimum if self.minimum is not None else -np.inf,
                self.maximum if self.maximum is not None else np.inf,
            )
        ]

    def __eq__(self, other):
        if not isinstance(other, FixedBound):
            return False
        if (
            (self.p == other.p)
            and (self.minimum == other.minimum)
            and (self.maximum == other.maximum)
        ):
            return True
        return False


#######


class ParametricConstraintList(MutableSequence):
    def __init__(self, init=None):
        self._cx = list()
        self.allow_dupes = False
        if init is not None:
            for i in init:
                if isinstance(i, ParametricConstraint):
                    self._cx.append(i)
                else:
                    raise TypeError(
                        f"members of {self.__class__.__name__} must be ParametricConstraint"
                    )

    def __getstate__(self):
        return self._cx, self.allow_dupes

    def __setstate__(self, state):
        self._cx, self.allow_dupes = state

    def set_instance(self, instance):
        self._instance = instance
        for i in self._cx:
            i.link_model(self._instance)

    def __fresh(self, instance):
        newself = ParametricConstraintList()
        newself._instance = instance
        setattr(instance, self.private_name, newself)
        return newself

    def __get__(self, instance, owner):
        if instance is None:
            return self
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        if newself is None:
            newself = self.__fresh(instance)
        return newself

    def __set__(self, instance, values):
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        if newself is None:
            newself = self.__fresh(instance)
        newself.__init__(values)
        try:
            newself._instance.mangle()
        except AttributeError:
            pass
        for i in newself._cx:
            i.link_model(newself._instance)

    def __delete__(self, instance):
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        newself.__init__()
        newself._instance = instance
        try:
            newself._instance.mangle()
        except AttributeError:
            pass

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_" + name

    def __getitem__(self, item):
        return self._cx[item]

    def __setitem__(self, key: int, value):
        if isinstance(value, str):
            value = interpret_contraint(value)
        if not isinstance(value, ParametricConstraint):
            raise TypeError("items must be of type ParametricConstraint")
        if self.allow_dupes or not self._is_duplicate(value):
            self._cx[key] = value
            try:
                self._instance.mangle()
            except AttributeError:
                pass
            self._cx[key].link_model(self._instance)

    def __delitem__(self, key):
        del self._cx[key]
        try:
            self._instance.mangle()
        except AttributeError:
            pass

    def __len__(self):
        return len(self._cx)

    def insert(self, index, value):
        if isinstance(value, str):
            value = interpret_contraint(value)
        if not isinstance(value, ParametricConstraint):
            raise TypeError("items must be of type ParametricConstraint")
        if self.allow_dupes or not self._is_duplicate(value):
            self._cx.insert(index, value)
            try:
                self._instance.mangle()
            except AttributeError:
                pass
            self._cx[index].link_model(self._instance)

    def __repr__(self):
        return repr(self._cx)

    def _is_duplicate(self, value):
        for i in self._cx:
            if i == value:
                return True
        return False

    def rescale(self, scale=None, index=None):
        if index is None:
            for i in self._cx:
                if scale is None:
                    i.link_model(self._instance, np.exp(np.random.uniform(-3, 3)))
                else:
                    i.link_model(self._instance, scale)
        else:
            if scale is None:
                self._cx[index].link_model(
                    self._instance, np.exp(np.random.uniform(-3, 3))
                )
            else:
                self._cx[index].link_model(self._instance, scale)


def asfloat(x):
    if "/" in x:
        x_ = x.split("/")
        return float(x_[0]) / float(x_[1])
    return float(x)


def interpret_contraint(c):
    numbr = r"\s*([-+]?\d*\.?\d*[eE]?[-+]?\d*|\d+\/\d+)\s*"
    token = r"\s*([\w#:\*&^%\$!@]+)\s*"

    n_2w = f"^{numbr}(<=|<){token}(<=|<){numbr}$"
    rx = re.match(n_2w, c)
    if rx:
        return FixedBound(
            rx.group(3), minimum=asfloat(rx.group(1)), maximum=asfloat(rx.group(5))
        )

    le_n = f"^{token}(<=|<){numbr}$"
    rx = re.match(le_n, c)
    if rx:
        return FixedBound(rx.group(1), maximum=asfloat(rx.group(3)))

    ge_n = f"^{token}(>=|>){numbr}$"
    rx = re.match(ge_n, c)
    if rx:
        return FixedBound(rx.group(1), minimum=asfloat(rx.group(3)))

    n_le = f"^{numbr}(<=|<){token}$"
    rx = re.match(n_le, c)
    if rx:
        return FixedBound(rx.group(3), minimum=asfloat(rx.group(1)))

    n_ge = f"^{numbr}(>=|>){token}$"
    rx = re.match(n_ge, c)
    if rx:
        return FixedBound(rx.group(3), maximum=asfloat(rx.group(1)))

    le = f"^{token}(<=|<){token}$"
    rx = re.match(le, c)
    if rx:
        return OrderingBound(rx.group(1), rx.group(3))

    ge = f"^{token}(>=|>){token}$"
    rx = re.match(ge, c)
    if rx:
        return OrderingBound(rx.group(3), rx.group(1))

    r_le = f"^{token}/{token}(<=|<){numbr}$"
    rx = re.match(r_le, c)
    if rx:
        return RatioBound(rx.group(1), rx.group(2), max_ratio=asfloat(rx.group(4)))

    r_ge = f"^{token}/{token}(>=|>){numbr}$"
    rx = re.match(r_ge, c)
    if rx:
        return RatioBound(rx.group(1), rx.group(2), min_ratio=asfloat(rx.group(4)))

    r_2w = f"^{numbr}(<=|<){token}/{token}(<=|<){numbr}$"
    rx = re.match(r_2w, c)
    if rx:
        return RatioBound(
            rx.group(3),
            rx.group(4),
            min_ratio=asfloat(rx.group(1)),
            max_ratio=asfloat(rx.group(6)),
        )

    raise ValueError(f"cannot interpret '{c}' as a constraint")



"""Provides a cache for requests.get and requests.post."""

from __future__ import annotations

import requests

cache_dir = None
memory = None


def set_cache_dir(location=None, compress=True, verbose=0, **kwargs):
    """
    Set up a cache directory for use with requests.

    Parameters
    ----------
    location: str or None or False
            The path of the base directory to use as a data store
            or None or False.  If None, a default directory is created
            using platformdirs.user_cache_dir.
            If False is given, no caching is done and
            the Memory object is completely transparent.

    compress: boolean, or integer, optional
            Whether to zip the stored data on disk. If an integer is
            given, it should be between 1 and 9, and sets the amount
            of compression.

    verbose: int, optional
            Verbosity flag, controls the debug messages that are issued
            as functions are evaluated.

    bytes_limit: int, optional
            Limit in bytes of the size of the cache.

    """
    global memory, cache_dir

    if location is None:
        import platformdirs  # imported here as this is an optional install

        location = platformdirs.user_cache_dir("cached_requests")

    if location is False:
        location = None

    import joblib  # imported here as this is an optional install

    memory = joblib.Memory(location, compress=compress, verbose=verbose, **kwargs)

    make_cache = (
        (requests, "get"),
        (requests, "post"),
    )

    for module, func_name in make_cache:
        try:
            func = getattr(module, f"_{func_name}_orig")
        except AttributeError:
            func = getattr(module, func_name)
            setattr(module, f"_{func_name}_orig", func)
        setattr(module, func_name, memory.cache(func))


set_cache_dir()



from __future__ import annotations

from enum import IntEnum

import numpy as np
import pandas as pd
import pytest
import xarray as xr

import sharrow


def test_simple_cat(tours_dataset: xr.Dataset):
    tree = sharrow.DataTree(tours=tours_dataset)

    assert all(tours_dataset.TourMode.cat.categories == ["Bus", "Car", "Walk"])

    expr = "tours.TourMode == 'Bus'"
    f = tree.setup_flow({expr: expr})
    a = f.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 1, 0, 0, 0]))

    tour_mode_bus = tree.get_expr(expr)
    assert all(tour_mode_bus == np.asarray([0, 1, 0, 0, 0]))


def test_2_level_tree_cat(
    tours_dataset: xr.Dataset,
    person_dataset: xr.Dataset,
):
    tree = sharrow.DataTree(tours=tours_dataset)
    tree.add_dataset("persons", person_dataset, "tours.person_id @ persons.person_id")

    assert all(tours_dataset.TourMode.cat.categories == ["Bus", "Car", "Walk"])

    expr = "tours.TourMode == 'Bus'"
    f = tree.setup_flow({expr: expr})
    a = f.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 1, 0, 0, 0]))

    tour_mode_bus = tree.get_expr(expr)
    assert all(tour_mode_bus == np.asarray([0, 1, 0, 0, 0]))

    work_mode_bus = tree.get_expr("WorkMode == 'Walk'")
    assert all(work_mode_bus == np.asarray([0, 0, 0, 0, 1]))

    work_mode_bus1 = tree.get_expr("persons.WorkMode == 'Walk'")
    assert all(work_mode_bus1 == np.asarray([0, 0, 0, 0, 1]))


def test_3_level_tree_cat(
    tours_dataset: xr.Dataset,
    person_dataset: xr.Dataset,
    household_dataset: xr.Dataset,
):
    tree = sharrow.DataTree(tours=tours_dataset)
    tree.add_dataset("persons", person_dataset, "tours.person_id @ persons.person_id")
    tree.add_dataset(
        "households", person_dataset, "persons.household_id @ households.household_id"
    )

    assert all(tours_dataset.TourMode.cat.categories == ["Bus", "Car", "Walk"])

    expr = "tours.TourMode == 'Bus'"
    f = tree.setup_flow({expr: expr})
    a = f.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 1, 0, 0, 0]))

    tour_mode_bus = tree.get_expr(expr)
    assert all(tour_mode_bus == np.asarray([0, 1, 0, 0, 0]))

    work_mode_bus = tree.get_expr("WorkMode == 'Walk'")
    assert all(work_mode_bus == np.asarray([0, 0, 0, 0, 1]))

    work_mode_bus1 = tree.get_expr("persons.WorkMode == 'Walk'")
    assert all(work_mode_bus1 == np.asarray([0, 0, 0, 0, 1]))


def test_rootless_tree_cat(
    tours_dataset: xr.Dataset,
    person_dataset: xr.Dataset,
    household_dataset: xr.Dataset,
):
    tree = sharrow.DataTree(tours=tours_dataset, root_node_name=False)
    tree.add_dataset("persons", person_dataset, "tours.person_id @ persons.person_id")
    tree.add_dataset(
        "households", person_dataset, "persons.household_id @ households.household_id"
    )

    assert all(tours_dataset.TourMode.cat.categories == ["Bus", "Car", "Walk"])

    expr = "tours.TourMode == 'Bus'"
    f = tree.setup_flow({expr: expr}, with_root_node_name="tours")
    a = f.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 1, 0, 0, 0]))


def test_int_enum_categorical():
    class TourMode(IntEnum):
        Car = 1
        Bus = 2
        Walk = 3

    df = pd.DataFrame(
        {
            "TourMode": ["Car", "Bus", "Car", "Car", "Walk"],
            "person_id": [441, 445, 552, 556, 934],
        },
        index=pd.Index([4411, 4451, 5521, 5561, 9341], name="tour_id"),
    )
    df["TourMode2"] = df["TourMode"].as_int_enum(TourMode)
    assert df["TourMode2"].dtype == "category"
    assert all(df["TourMode2"].cat.categories == ["_0", "Car", "Bus", "Walk"])
    assert all(df["TourMode2"].cat.codes == [1, 2, 1, 1, 3])


def test_missing_categorical():
    df = pd.DataFrame(
        {
            "TourMode": ["Car", "Bus", "Car", "Car", "Walk", np.nan],
            "person_id": [441, 445, 552, 556, 934, 998],
        },
        index=pd.Index([4411, 4451, 5521, 5561, 9341, 9981], name="tour_id"),
    )
    df["TourMode2"] = df["TourMode"].astype(pd.CategoricalDtype(["Car", "Bus", "Walk"]))
    assert df["TourMode2"].dtype == "category"
    assert all(df["TourMode2"].cat.categories == ["Car", "Bus", "Walk"])
    assert all(df["TourMode2"].cat.codes == [0, 1, 0, 0, 2, -1])

    tree = sharrow.DataTree(df=df, root_node_name=False)

    expr = "df.TourMode2 == 'Bus'"
    f = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = f.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 1, 0, 0, 0, 0]))

    expr = "df.TourMode2.isna()"
    f2 = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = f2.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 0, 0, 0, 0, 1]))

    expr = "df.TourMode2 == 'Walk'"
    f3 = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = f3.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 0, 0, 0, 1, 0]))

    expr = "'Walk' == df.TourMode2"
    f4 = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = f4.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 0, 0, 0, 1, 0]))

    expr = "df.TourMode2 == 'BAD'"
    with pytest.warns(UserWarning):
        f5 = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = f5.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 0, 0, 0, 0, 0]))

    expr = "'BAD' == df.TourMode2"
    with pytest.warns(UserWarning):
        f6 = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = f6.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 0, 0, 0, 0, 0]))

    expr = "df.TourMode2 != 'Bus'"
    f7 = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = f7.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([1, 0, 1, 1, 1, 1]))

    expr = "df.TourMode2 != 'BAD'"
    with pytest.warns(UserWarning):
        f8 = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = f8.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([1, 1, 1, 1, 1, 1]))

    expr = "'BAD' != df.TourMode2"
    with pytest.warns(UserWarning):
        f9 = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = f9.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([1, 1, 1, 1, 1, 1]))

    expr = "(df.TourMode2 == 'BAD') * 2"
    with pytest.warns(UserWarning):
        fA = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = fA.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 0, 0, 0, 0, 0]))

    expr = "(df.TourMode2 == 'BAD') * 2.2"
    with pytest.warns(UserWarning):
        fB = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = fB.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([0, 0, 0, 0, 0, 0]))

    expr = "np.exp(df.TourMode2 == 'BAD') * 2.2"
    with pytest.warns(UserWarning):
        fC = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = fC.load_dataarray(dtype=np.float32)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([2.2, 2.2, 2.2, 2.2, 2.2, 2.2], dtype=np.float32))

    expr = "(df.TourMode2 != 'BAD') * 2"
    with pytest.warns(UserWarning):
        fD = tree.setup_flow({expr: expr}, with_root_node_name="df")
    a = fD.load_dataarray(dtype=np.int8)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([2, 2, 2, 2, 2, 2]))


def test_categorical_indexing(tours_dataset: xr.Dataset, skims_dataset: xr.Dataset):
    tree = sharrow.DataTree(tours=tours_dataset)
    tree.add_dataset(
        "od_skims",
        skims_dataset,
        [
            "tours.origin @ od_skims.otaz",
            "tours.destination @ od_skims.dtaz",
            "tours.time_period @ od_skims.timeperiod",
        ],
    )

    expr = "od_skims.cartime"
    f = tree.setup_flow({expr: expr})
    a = f.load_dataarray(dtype=np.float32)
    a = a.isel(expressions=0)
    assert all(a == np.asarray([22.2, 6.6, 99.9, 11.1, 8.8], dtype=np.float32))

    skims_dataset_bad = (
        skims_dataset["cartime"].sel(timeperiod=["MD", "AM"]).to_dataset()
    )
    with pytest.raises(ValueError, match="categoricals have different categories"):
        tree.replace_datasets(od_skims=skims_dataset_bad)

    # test with a missing value ...
    tours_dataset_bad = tours_dataset.copy(deep=True)
    tours_dataset_bad["time_period"].loc[dict(tour_id=4411)] = -1

    # test with a missing value when creating a tree
    bad_tree = sharrow.DataTree(tours=tours_dataset_bad)
    bad_tree.add_dataset(
        "od_skims",
        skims_dataset,
        [
            "tours.origin @ od_skims.otaz",
            "tours.destination @ od_skims.dtaz",
            "tours.time_period @ od_skims.timeperiod",
        ],
    )
    with pytest.raises(ValueError, match="detected missing values"):
        bad_tree.setup_flow({expr: expr})

    # test with a missing value when replacing datasets
    with pytest.raises(ValueError, match="detected missing values"):
        tree.replace_datasets(tours=tours_dataset_bad)


def test_bad_categorical_indexing(tours_dataset: xr.Dataset, skims_dataset: xr.Dataset):
    tree = sharrow.DataTree(tours=tours_dataset)
    tree.add_dataset(
        "od_skims",
        skims_dataset,
        [
            "tours.origin @ od_skims.otaz",
            "tours.destination @ od_skims.dtaz",
            "tours.time_period_alt @ od_skims.timeperiod",
        ],
    )

    expr = "od_skims.cartime"
    with pytest.raises(ValueError, match="categoricals have different categories"):
        # this fails because `time_period_alt` is intentionally constructed backwards
        # and does not match the `od_skims.timeperiod` categories
        tree.setup_flow({expr: expr})



try:
    from ._version import __version__, __version_tuple__
except ImportError:
    # Package is not "installed", parse git tag at runtime
    from importlib.metadata import PackageNotFoundError, version

    try:
        __version__ = version(__package__)
    except PackageNotFoundError:
        # package is not installed
        __version__ = "999.999"

    __version_tuple__ = __version__.split(".")



from __future__ import annotations

import logging
from numbers import Number
from typing import TYPE_CHECKING, Literal

import numpy as np
import xarray as xr

if TYPE_CHECKING:
    from larch import BaseModel


def new_name(existing_names):
    n = 1
    while f"model{n:05d}" in existing_names:
        n += 1
    return f"model{n:05d}"


def empty(dtype, index_name):
    return xr.DataArray(np.empty(0, dtype=dtype), dims=index_name)


class ParameterBucket:
    index_name = "param_name"
    _data_types = {
        "value": np.float32,
        "initvalue": np.float32,
        "nullvalue": np.float32,
        "minimum": np.float32,
        "maximum": np.float32,
        "holdfast": np.int8,
    }

    def __init__(self, *models, **kmodels):
        self._models = {}
        self._params = xr.Dataset(
            {
                "value": empty(np.float32, self.index_name),
                "initvalue": empty(np.float32, self.index_name),
                "nullvalue": empty(np.float32, self.index_name),
                "minimum": empty(np.float32, self.index_name),
                "maximum": empty(np.float32, self.index_name),
                "holdfast": empty(np.int8, self.index_name),
            },
            coords={self.index_name: empty(np.object_, self.index_name)},
        )
        self._fill_values = {
            "value": 0.0,
            "initvalue": 0.0,
            "nullvalue": 0.0,
            "minimum": -np.inf,
            "maximum": np.inf,
            "holdfast": np.int8(0),
        }
        for k, m in kmodels.items():
            self.attach_model(m, k, agg=False)
        for m in models:
            if isinstance(m, dict):
                for k_, m_ in m.items():
                    self.attach_model(m_, k_, agg=False)
            else:
                self.attach_model(m, agg=False)
        self._aggregate_parameters()

    def attach_model(
        self,
        model: BaseModel,
        name: str | None = None,
        agg: bool = True,
        unmangle: bool | Literal["structure":] = True,
    ):
        """
        Attach a model to the bucket.

        Parameters
        ----------
        model : BaseModel
            The model to attach.
        name : str, optional
            The name to attach the model as.  If None, the model's ident
            attribute is used.  If the model has no ident attribute, a new
            name is generated.
        agg : bool
            If True, aggregate the parameters of the model with the bucket.
        unmangle : bool or str
            If True, unmangle the model.  If "structure", only unmangle the
            structure of the model.
        """
        if name is None:
            try:
                name = model.ident
            except AttributeError:
                pass
        else:
            self.ident = name
        if name is None:
            raise ValueError("missing name")
            # name = new_name(self._models.keys())
        if unmangle == "structure":
            model.unmangle(structure_only=True)
        elif unmangle:
            model.unmangle()
        self._models[name] = model

        # collect parameters from the incoming model to be attached, to be able
        # to update the bucket's parameters if they are not already present
        move_values = {}
        for k in model._parameter_bucket.pnames:
            if k not in self._params[self.index_name]:
                move_values[k] = model._parameter_bucket.parameters.sel(
                    {self.index_name: k}
                ).copy(deep=True)
        model._parameter_bucket = self
        if agg:
            if move_values:
                self.combine_parameters(
                    xr.concat(move_values.values(), dim="param_name")
                )
            self._aggregate_parameters()

    def detach_model(self, model):
        name = getattr(model, "_ident", None)
        # if getattr(self._models[name], "_parameter_bucket", None) is self:
        #     del self._models[name]._parameter_bucket
        if name is None:
            return
        elif self._models.get(name, None) is model:
            del self._models[name]
        if getattr(model, "_parameter_bucket", None) is self:
            del model._parameter_bucket

    def rename_model(self, oldname, newname):
        if oldname == newname:
            return
        try:
            m = self._models[oldname]
        except KeyError:
            print("keys:", list(self._models.keys()))
            raise
        if newname in self._models:
            raise ValueError("name exists")
        self._models[newname] = m
        del self._models[oldname]

    def _aggregate_parameters(self):
        should_mangle = False
        all_names = set()
        for m in self._models.values():
            all_names |= set(m.pnames)
        new_params = self._params.reindex(
            {self.index_name: sorted(all_names)}, fill_value=self._fill_values
        )
        if new_params[self.index_name].size != self._params[
            self.index_name
        ].size or any(new_params[self.index_name] != self._params[self.index_name]):
            should_mangle = True
        self._params = new_params
        if should_mangle:
            self.mangle()

    def _preferred_data_type(self, key):
        t = self._data_types.get(key, None)
        if t is None:
            if key in self._params:
                t = self._params[key].dtype
        if t is None:
            t = np.object_
        return t

    def _clean_dtypes(self, new_params):
        for k in new_params:
            new_params = new_params.assign(
                {k: new_params[k].astype(self._preferred_data_type(k))}
            )
        return new_params

    def update_parameters(self, other):
        if isinstance(other, ParameterBucket):
            other = other._params
        if isinstance(other, xr.Dataset):
            if "param_name_2" in other.sizes:
                other = other.drop_dims("param_name_2")  # TODO, keep 2nd dim
            if len(other.sizes) != 1:
                raise ValueError("expected parameters to be a 1-d vector")
            other_dim = iter(other.sizes).__next__()
            joint_names = set(self.pnames) | set(other.coords[other_dim].data)
        else:
            joint_names = set(self.pnames) | set(other.pnames)
        should_mangle = False
        new_params = self._params.reindex({self.index_name: sorted(joint_names)})
        filling = [i for i in new_params if i in other]
        try:
            new_params = new_params.fillna(other[filling])
        except ValueError as err:
            logging.getLogger(__name__).error(f"other={other}")
            logging.getLogger(__name__).exception(f"err={err}")
        if new_params[self.index_name].size != self._params[
            self.index_name
        ].size or any(new_params[self.index_name] != self._params[self.index_name]):
            should_mangle = True
        new_params = self._clean_dtypes(new_params)
        self._params = new_params
        if should_mangle:
            self.mangle()

    def combine_parameters(self, other: xr.Dataset | ParameterBucket):
        """
        Combine the parameters with another set of parameters.

        Values in the current parameter set are replaced by values in the
        other parameter set.
        """
        if isinstance(other, ParameterBucket):
            other = other._params
        if isinstance(other, xr.Dataset):
            if "param_name_2" in other.sizes:
                other = other.drop_dims("param_name_2")
        if len(other.sizes) != 1:
            raise ValueError("expected parameters to be a 1-d vector")
        if not isinstance(other, xr.Dataset):
            raise TypeError("expected other to be a Dataset")
        should_mangle = False
        n_existing_params = self._params.sizes[self.index_name]
        self._params = other.combine_first(self._params)
        if self._params.sizes[self.index_name] != n_existing_params:
            should_mangle = True
        if should_mangle:
            self.mangle()

    def add_parameters(self, names, fill_values=None):
        should_mangle = False
        joint_names = set(self.pnames) | set(names)
        fills = self._fill_values.copy()
        if fill_values is not None:
            fills.update(fill_values)
        new_params = self._params.reindex(
            {self.index_name: sorted(joint_names)}, fill_value=fills
        )
        if new_params[self.index_name].size != self._params[
            self.index_name
        ].size or any(new_params[self.index_name] != self._params[self.index_name]):
            should_mangle = True
        new_params = self._clean_dtypes(new_params)
        self._params = new_params
        if should_mangle:
            self.mangle()

    def add_array(self, name, values):
        for j in range(values.ndim):
            if values.shape[j] != self.n_params:
                raise ValueError(f"cannot add array, unexpected shape {values.shape}")
        if values.ndim == 1:
            dims = [self.index_name]
        else:
            dims = [self.index_name] + [
                f"{self.index_name}_{j+2}" for j in range(values.ndim - 1)
            ]
        self._params = self._params.assign(
            {name: xr.DataArray(np.asarray(values), dims=dims)}
        )

    def add_parameter_array(self, name, values):
        return self.add_array(name, values)

    def mangle(self, data=True, structure=True):
        for m in self._models.values():
            m.mangle(data=data, structure=structure)
            m._ensure_names(self.pnames)

    def unmangle(self, *args, **kwargs):
        for m in self._models.values():
            m.unmangle(*args, **kwargs)

    def _scan_all_ensure_names(self):
        for m in self._models.values():
            m._scan_all_ensure_names()

    @property
    def parameters(self):
        return self._params

    @property
    def n_params(self):
        return self._params.sizes[self.index_name]

    @property
    def pvals(self):
        return self._params["value"].to_numpy()

    @pvals.setter
    def pvals(self, x):
        if isinstance(x, str):
            if x in self._params:
                x = self._params[x].to_numpy()
            elif f"{x}value" in self._params:
                x = self._params[f"{x}value"].to_numpy()
            elif f"{x}imum" in self._params:
                x = self._params[f"{x}imum"].to_numpy()
            else:
                raise ValueError(f"unknown value set {x}")
        if isinstance(x, dict):
            candidates = (
                xr.DataArray(
                    np.asarray(list(x.values())),
                    dims=self.index_name,
                    coords={self.index_name: np.asarray(list(x.keys()))},
                )
                .reindex({self.index_name: self.pnames})
                .fillna(self.pvals)
            )
            x = np.where(
                self._params["holdfast"].to_numpy(),
                self._params["value"].to_numpy(),
                candidates.to_numpy(),
            )
        elif isinstance(x, Number):
            candidates = xr.full_like(self._params["value"], x)
            x = np.where(
                self._params["holdfast"].to_numpy(),
                self._params["value"].to_numpy(),
                candidates.to_numpy(),
            )
        else:
            # we are getting an array
            x = np.asanyarray(x).reshape(-1)
            assert x.size == self._params["value"].size
            # do not allow changing holdfast values
            x = np.where(
                self._params["holdfast"].to_numpy(),
                self._params["value"].to_numpy(),
                x,
            )
        self._params = self._params.assign(
            {"value": xr.DataArray(x, dims=self._params["value"].dims)}
        )

    @property
    def pinitvals(self):
        return self._params["initvalue"].to_numpy()

    @pinitvals.setter
    def pinitvals(self, x):
        if isinstance(x, str):
            raise ValueError("cannot set initvals with a string")
        if isinstance(x, dict):
            candidates = (
                xr.DataArray(
                    np.asarray(list(x.values())),
                    dims=self.index_name,
                    coords={self.index_name: np.asarray(list(x.keys()))},
                )
                .reindex({self.index_name: self.pnames})
                .fillna(self.pvals)
            )
            x = np.where(
                self._params["holdfast"].to_numpy(),
                self._params["initvalue"].to_numpy(),
                candidates.to_numpy(),
            )
        elif isinstance(x, Number):
            candidates = xr.full_like(self._params["value"], x)
            x = np.where(
                self._params["holdfast"].to_numpy(),
                self._params["initvalue"].to_numpy(),
                candidates.to_numpy(),
            )
        else:
            # we are getting an array
            x = np.asanyarray(x).reshape(-1)
            assert x.size == self._params["initvalue"].size
            # do not allow changing holdfast values
            x = np.where(
                self._params["holdfast"].to_numpy(),
                self._params["initvalue"].to_numpy(),
                x,
            )
        self._params = self._params.assign(
            {"initvalue": xr.DataArray(x, dims=self._params["initvalue"].dims)}
        )

    @property
    def pholdfast(self):
        return self._params["holdfast"].to_numpy()

    @property
    def pnullvals(self):
        return self._params["nullvalue"].to_numpy()

    @pholdfast.setter
    def pholdfast(self, x):
        if isinstance(x, dict):
            x = (
                xr.DataArray(
                    np.asarray(list(x.values())),
                    dims=self.index_name,
                    coords={self.index_name: np.asarray(list(x.keys()))},
                )
                .reindex({self.index_name: self.pnames})
                .fillna(self.pholdfast)
                .astype(np.int8)
            )
        self._params = self._params.assign(
            {"holdfast": xr.DataArray(x, dims=self._params["holdfast"].dims)}
        )

    @pnullvals.setter
    def pnullvals(self, x):
        if isinstance(x, dict):
            x = (
                xr.DataArray(
                    np.asarray(list(x.values())),
                    dims=self.index_name,
                    coords={self.index_name: np.asarray(list(x.keys()))},
                )
                .reindex({self.index_name: self.pnames})
                .fillna(self.pnullvals)
            )
        self._params = self._params.assign(
            {"nullvalue": xr.DataArray(x, dims=self._params["nullvalue"].dims)}
        )

    @property
    def pmaximum(self):
        return self._params["maximum"].to_numpy()

    @pmaximum.setter
    def pmaximum(self, x):
        if isinstance(x, dict):
            candidates = (
                xr.DataArray(
                    np.asarray(list(x.values())),
                    dims=self.index_name,
                    coords={self.index_name: np.asarray(list(x.keys()))},
                )
                .reindex({self.index_name: self.pnames})
                .fillna(self.pmaximum)
            )
            x = np.where(
                self._params["holdfast"].to_numpy(),
                self._params["maximum"].to_numpy(),
                candidates.to_numpy(),
            )
        if isinstance(x, Number):
            candidates = xr.full_like(self._params["value"], x)
            x = np.where(
                self._params["holdfast"].to_numpy(),
                self._params["maximum"].to_numpy(),
                candidates.to_numpy(),
            )
        self._params = self._params.assign(
            {"maximum": xr.DataArray(x, dims=self._params["maximum"].dims)}
        )

    @property
    def pminimum(self):
        return self._params["minimum"].to_numpy()

    @pminimum.setter
    def pminimum(self, x):
        if isinstance(x, dict):
            candidates = (
                xr.DataArray(
                    np.asarray(list(x.values())),
                    dims=self.index_name,
                    coords={self.index_name: np.asarray(list(x.keys()))},
                )
                .reindex({self.index_name: self.pnames})
                .fillna(self.pminimum)
            )
            x = np.where(
                self._params["holdfast"].to_numpy(),
                self._params["minimum"].to_numpy(),
                candidates.to_numpy(),
            )
        if isinstance(x, Number):
            candidates = xr.full_like(self._params["value"], x)
            x = np.where(
                self._params["holdfast"].to_numpy(),
                self._params["minimum"].to_numpy(),
                candidates.to_numpy(),
            )
        self._params = self._params.assign(
            {"minimum": xr.DataArray(x, dims=self._params["minimum"].dims)}
        )

    @property
    def pbounds(self):
        """scipy.optimize.Bounds : A copy of the current min-max bounds of the parameters."""
        self.unmangle()
        from scipy.optimize import Bounds

        return Bounds(
            self.pminimum,
            self.pmaximum,
        )

    def set_cap(self, cap=25):
        """
        Set limiting values for one or more parameters.

        Parameters
        ----------
        cap : numeric, default 25.0
            Set a global limit on parameters.  The maximum has a ceiling
            at this value, and the minimum a floor at the negative of this, unless
            the existing bounds are entirely outside this range.
        """
        # propose capping all parameters
        minimum_ = np.maximum(self.pminimum, -cap)
        maximum_ = np.minimum(self.pmaximum, cap)

        # anywhere the minmax range was fully collapsed, restore the originals
        minimum_ = np.where(minimum_ <= maximum_, minimum_, self.pminimum)
        maximum_ = np.where(minimum_ <= maximum_, maximum_, self.pmaximum)

        # set any NaN bounds to the cap
        minimum_ = np.where(np.isnan(minimum_), -cap, minimum_)
        maximum_ = np.where(np.isnan(maximum_), cap, maximum_)

        # restore the originals for holdfast parameters
        minimum_ = np.where(self.pholdfast, self.pminimum, minimum_)
        maximum_ = np.where(self.pholdfast, self.pmaximum, maximum_)

        # write out new bounds
        self._params = self._params.assign(
            {
                "minimum": xr.DataArray(minimum_, dims=self._params["value"].dims),
                "maximum": xr.DataArray(maximum_, dims=self._params["value"].dims),
            }
        )

    @property
    def pnames(self):
        return self._params[self.index_name].to_numpy()

    def get_param_loc(self, name) -> int:
        """
        Get the position of a named parameter.

        Parameters
        ----------
        name : str
            Name of parameter to find

        Returns
        -------
        int
            Position of parameter in array

        Raises
        ------
        KeyError
            The parameter is not found
        """
        name = str(name)
        return self._params.indexes[self.index_name].get_loc(name)

    @property
    def pstderr(self):
        if "std_err" in self._params:
            return self._params["std_err"].to_numpy()
        else:
            return np.full_like(self.pvals, np.nan)

    @pstderr.setter
    def pstderr(self, x):
        self._params = self._params.assign(
            {"std_err": xr.DataArray(x, dims=self._params["value"].dims)}
        )

    def __getitem__(self, item):
        if item in self._models:
            return self._models[item]
        raise KeyError

    def _assign(self, key, param_name, value):
        self._params = self._params.assign(
            {
                key: self._params[key].where(
                    self._params[self.index_name] != param_name, value
                )
            }
        )

    def lock(self, values=None, **kwargs) -> None:
        """
        Lock the values of one or more parameters.

        Parameters
        ----------
        values : dict, optional
            Dictionary of parameters to lock.  The keys are parameter names
            and the values are the values to set as locked.
        kwargs : dict
            Additional parameters to lock.
        """
        if values is None:
            values = {}
        values.update(kwargs)
        for k, v in values.items():
            self._assign("value", k, v)
            self._assign("initvalue", k, v)
            self._assign("holdfast", k, 1)

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_" + name

    def __get__(self, instance, owner=None):
        """
        Get attribute from the instance.

        Parameters
        ----------
        instance : Any
            Instance of parent class that has `self` as a member.
        instancetype : class
            Class of `instance`.
        """
        if instance is None:
            return self
        bucket = getattr(instance, self.private_name, None)
        return bucket

    def __set__(self, instance, value):
        """
        Set attribute on the instance.

        Parameters
        ----------
        instance : Any
            Instance of parent class that has `self` as a member.
        value : ParameterBucket
            The bucket being assigned.
        """
        if not isinstance(value, ParameterBucket):
            raise TypeError(f"{self.name} must be a ParameterBucket")
        existing = getattr(instance, self.private_name, None)
        if existing is value:
            return
        if existing is not None:
            existing.detach_model(instance)
        # _name_in_parameter_bucket = getattr(instance, "_name_in_parameter_bucket", None)
        setattr(instance, self.private_name, value)
        instance.mangle()
        value.attach_model(instance, agg=False, unmangle="structure")
        if existing is not None:
            value.update_parameters(existing)
        instance.mangle()

    def __delete__(self, instance):
        existing = getattr(instance, self.private_name, None)
        setattr(instance, self.private_name, None)
        if existing is not None:
            existing.detach_model(instance)
        instance.mangle()

    def pretty_table(self, name_width=25):
        from rich.table import Table

        params = self._params
        rich_table = Table()
        idx_name, n = next(iter(params.sizes.items()))
        for column in params.variables:
            rich_table.add_column(
                str(column), width=name_width if column == self.index_name else None
            )
        for i in range(n):
            row = [str(params[j].data[i]) for j in params.variables]
            rich_table.add_row(*row)
        return rich_table
